#!/usr/bin/python
#
# Pyring 
#
# copyright Angus Ainslie <angus.ainslie@gmail.com>

import struct
import socket

import pygtk
import gtk
import gtk.glade
import hashlib
import Numeric
import pyDes
import binascii 
import string
import os
import re
import readDB

try:
    import hildon
except:
    hasHildon = False
else:
    hasHildon = True

from xml.sax import make_parser
from xml.sax.handler import feature_namespaces

class Main:
    def __init__(self):
        # do some initalization  
        if hasHildon :
            print "Using hildon UI"
            self.gladeFile = "/usr/lib/pyring/gui.glade"
        else :
            print "Using Standard UI"
            self.gladeFile = "gui.glade"

        self.wTree = gtk.glade.XML( self.gladeFile, "mainwindow" )

        signals = { 
            "on_quit_activate" : self.OnQuit,
            "on_save_activate" : self.OnSave,
            "on_delete_activate" : self.OnRowDelete,
            "on_new_activate" : self.OnRowNew,
            "on_about_activate" : self.OnAbout,
            "on_aboutdialog_close" : self.OnAboutClose,
            "on_change_password_activate" : self.OnChangePassword,
            "on_mainwindow_delete_event" : self.OnDestroy,
            "on_import_activate" : self.OnImport,
            "on_nameView_row_activated" : self.OnRowActivated,
            "on_nameView_select_cursor_row" : self.OnRowActivated,
            "on_nameView_sort" : self.NameViewSort,
            "on_name_text_changed" : self.OnNameTextChanged,
            "on_account_text_changed" : self.OnAccountTextChanged,
            "on_password_text_changed" : self.OnPasswordTextChanged,
            "on_note_text_changed" : self.OnNoteTextChanged,
            "on_main_update_btn_activate" : self.OnMainUpdateBtn,
            "on_main_cancel_btn_activate" : self.OnMainCancelBtn,
        }

        self.wTree.signal_autoconnect(signals)

        #Here are some variables that can be reused later
        self.saltSize = 4
        self.salt = None
        self.hashSize = 16
        self.cName = 0
        self.cUsername = 1
        self.hasChanged = False

        self.sName = "Name"
        self.sUsername = "Username"

        self.userDir = os.path.expanduser( '~/MyDocs' )
        self.confDir = os.path.join( self.userDir, '.pyring' )
        self.confFile = os.path.join( self.confDir, "pyring.conf" )
        self.dataFile = os.path.join( self.confDir, "pyringDB.xml" )

        self.pwDigest = None
        self.hash = ''
        self.records = []
        self.recordChanged = 0

        if not os.path.isdir( self.confDir ) :
            print "Creating config directory :", self.confDir
            os.mkdir( self.confDir )

        if not os.path.isfile( self.confFile ) :
            print "Creating config file :", self.confFile
            f = open( self.confFile, 'wb' )
            f.write( "config file\n" )
            f.close

        
        #Get the treeView from the widget Tree
        self.nameView = self.wTree.get_widget("nameView")
        
        self.AddListColumn(self.sName, self.cName, True)
        self.AddListColumn(self.sUsername, self.cUsername, False)

        self.nameList = gtk.ListStore(str, str)
        self.nameList.connect("sort-column-changed", self.NameViewSort)
        self.nameView.set_model(self.nameList)

        if hasHildon == False :
            self.window = self.wTree.get_widget("mainwindow")
        else:
            self.app = hildon.Program()
            vMain = self.wTree.get_widget("vMain")
            self.window = hildon.Window()
            self.window.set_title('Pyring')
            self.window.connect("destroy", self.OnQuit)
            self.app.add_window(self.window)
 
            vMain.reparent(self.window)
#            self.pyring.destroy()

            menu = gtk.Menu()
            mainMenu = self.wTree.get_widget("mainmenu")
            for child in mainMenu.get_children():
                child.reparent(menu)
            self.window.set_menu(menu)

            mainMenu.destroy()

        self.window.show_all()

        # the list needs to be realized before it can be sorted
        if os.path.exists( self.dataFile ) :
            self.readPyringDB( self.dataFile )
            self.NameViewSort( self.nameList )
            for record in self.records : 
                self.nameList.append( [record[0], ''] )
 
    def OnChangePassword(self, widget ):
        password = self.GetPassword( "Old Password" )
        if self.pwDigest == None :
            if not self.CheckPalmKeyringPasswd( password, True ) :
                return False
        else :
            if not self.CheckPalmKeyringPasswd( password, False ) :
                return False

        password1 = self.GetPassword( "New Password" )
        password2 = self.GetPassword( "Again" )

        if password1 != password2 : 
            show_error_dialog("New passords don't match" )
            return False

        password.zfill( len( password ))
        password2.zfill( len( password2 ))

        m = hashlib.md5()
        m.update( self.salt )
        m.update( password1 )
        # pad with zeros
        m.update( ''.rjust( 64 - len( self.salt ) - len( password1 ), '\0' ) )
            
        self.hash = m.digest()

        m = hashlib.md5( password1 )
        newDigest = m.digest()
        
        password1.zfill( len( password1 ))

        record = ''

        for i in range( len( self.records )) :
            record = self.DecodeRecord( self.records[i][1], self.pwDigest )
            record = string.join( record, '\0' )
            #print "record : ", record.encode( "hex" )
            self.records[i][1] = self.EncodeRecord( record, newDigest )
    
            self.hasChanged = True
        self.pwDigest = newDigest

        return True

    def OnImport(self, widget ):
        if hasHildon :
            dialog = hildon.FileChooserDialog(self.window, 
                                           gtk.FILE_CHOOSER_ACTION_OPEN);
        else :
            dialog = gtk.FileChooserDialog("Import ..",
                                       None,
                                       gtk.FILE_CHOOSER_ACTION_OPEN,
                                       (gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL,
                                       gtk.STOCK_OPEN, gtk.RESPONSE_OK))
        dialog.set_default_response(gtk.RESPONSE_OK)

        filter = gtk.FileFilter()
        filter.set_name("Palm Database File")
        filter.add_pattern("*.pdb")

        dialog.add_filter(filter)
        dialog.show()
        while 1:
            response = dialog.run()
            if response == gtk.RESPONSE_OK:
                name = dialog.get_filename()
                if name != None :
                    print name
                    
                    keyFile = open( name, "rb")
                    dialog.destroy()
                        
                    self.hasChanged = True

                    return self.ReadPalmKeyringDB( keyFile )
            
                else:
                    dialog.destroy()
                    return False

            elif response == gtk.RESPONSE_CANCEL:
                dialog.destroy()
                return False

    def AlphaOnly( self, text ) :
        """list views have a funny way of sorting on on the alpha chars in the 
        list - This tries to do the same thing"""
        ret = ''
        a = []

        r = re.compile(r'\w+')
        a = r.findall( str( text ).lower() )

        ret = string.join(a, '' )

        #print "a : ", a, " ret: ", ret
        return ret

    def NameViewSort( self, list ):
        col = self.nameView.get_column( 0 )
        order = col.get_sort_order()

        if order == gtk.SORT_ASCENDING :
            self.records.sort( self.backward )
        else:
            self.records.sort( self.forward )

    def forward( self, a, b, ) :
        strip_a = self.AlphaOnly( a )
        strip_b = self.AlphaOnly( b )
        return cmp( strip_a, strip_b )

    def backward( self, a, b ) :
        strip_a = self.AlphaOnly( a )
        strip_b = self.AlphaOnly( b )
        return cmp( strip_b, strip_a )
    
    def readPyringDB( self, fileName ) :
        f = open( fileName, "rb")
        
        if f == None :
            return

        versionBytes = []
        hashBytes = []
        dh = readDB.ReadPyringXML( versionBytes, hashBytes, self.records )

        parser = make_parser()

        parser.setFeature(feature_namespaces, 0)
        parser.setContentHandler(dh)

        parser.parse( f )

        self.salt = hashBytes[0][0:self.saltSize]
        self.hash = hashBytes[0][self.saltSize:]
        #print "hashBytes : ", hashBytes
        #print "salt : ", self.salt.encode( "hex" )
        #print "hash : ", self.hash.encode( "hex" )
        
        f.close()

        #for record in self.records :
        #    print "Name : " , record[0]
        
    def writePyringDB( self, fileName ) :
        if os.path.exists( fileName ) :
            os.rename( fileName, fileName + '.old' )

        if self.records == [] :
            return

        f = open( fileName, "wb")

        f.write( "<records>\n" )
        f.write( "\t<version>" + '1.0' + "</version>\n" )
        str = binascii.b2a_hex( self.salt + self.hash )
        f.write( "\t<hash>" +str + "</hash>\n" )
        for record in self.records :
            f.write( "\t<record>\n" )
            str = binascii.b2a_hex( record[0] )            
            f.write( "\t\t<name>" + str + "</name>\n" )
            f.write( "\t\t<data>" )
            str = binascii.b2a_hex( record[1] )
            f.write( str )
            f.write( "</data>\n" )
            f.write( "\t</record>\n" )
        f.write( "</records>\n" )
                 
        f.close()

        #for record in self.records :
        #    print "Name : " , record[0]

    def CheckCurrentRow( self, validate ) :
        if not self.CheckPassword :
            return
            
        if self.recordChanged == 1 :
            self.recordChanged = 0
            if self.curRecord != None :
                if validate :
                    save = self.show_warning_dialog( "Apply Changes" )
                    if save == gtk.RESPONSE_CANCEL :
                        return False
                
                self.UpdateRecordFromRow()
                return True

        return False

    def OnQuit(self,widget) :
        self.CheckCurrentRow( True )

        if self.hasChanged :
            save = self.show_warning_dialog( "Save Changes ?" )
            if save == gtk.RESPONSE_OK :
                self.writePyringDB( self.dataFile )
        gtk.main_quit()

    def OnSave(self,widget) :
        self.ClearButtons()
        self.CheckCurrentRow( False )

        if self.hasChanged :
            self.hasChanged = False
            self.writePyringDB( self.dataFile )

    def OnDestroy(self,widget , arg ) :
        self.OnQuit( widget )

    def OnAboutClose( self, arg ) :
        self.aboutDlg.destroy()

    def OnAbout( self, arg ) :
        print "Arg : ", arg 

        aboutTree = gtk.glade.XML( self.gladeFile, "aboutdialog" )

        aboutTree.get_widget("aboutdialog").run()
        #self.aboutDlg.run()

    def GetPassword( self, title = 'Password' ) :
        passTree = gtk.glade.XML( self.gladeFile, "passwd_dlg" )

        passwdDlg = passTree.get_widget("passwd_dlg")
        passwdDlg.set_title( title )
        response = passwdDlg.run()
        passwd_entry = passTree.get_widget("passwd_entry")
        
        password = ''

        if response == gtk.RESPONSE_OK:
            password = passwd_entry.get_text()
            passwdDlg.destroy()
            if not password:
                return None
        else :
            passwdDlg.destroy()
        
        return password

    def OnNameTextChanged( self, event ) :
        self.recordChanged = 1
        self.ActivateButtons()

    def OnAccountTextChanged( self, event ) :
        self.recordChanged = 1
        self.ActivateButtons()

    def OnPasswordTextChanged( self, event ) :
        self.recordChanged = 1
        self.ActivateButtons()

    def OnNoteTextChanged( self, one, two ) :
        self.recordChanged = 1
        self.ActivateButtons()

    def OnMainCancelBtn( self, event ) :
        if not self.CheckPassword() :
            return
            
        record = self.DecodeRecord( self.records[self.curRecord][1]
                                    , self.pwDigest )
        #print "record : ", record
        self.SetFields( self.records[self.curRecord][0], record[0],
                            record[1], record[2] )
        self.recordChanged = 0
        self.ClearButtons()

    def CheckPassword( self ) :
        if self.pwDigest == None :
            password = self.GetPassword()
            if not self.CheckPalmKeyringPasswd( password ) :
                return False
            
        return True
        
    def UpdateRecordFromRow( self ) :
        if self.curRecord == None :
            print "Can't update None "
            return

        #print "Cur :", self.curRecord

        self.hasChanged = True
        fld = self.wTree.get_widget("name_text")
        string = fld.get_text()
        self.records[self.curRecord][0] = string

        fld = self.wTree.get_widget("account_text")
        string = fld.get_text()
        fld = self.wTree.get_widget("password_text")
        string = string + '\0' +  fld.get_text()
        fld = self.wTree.get_widget("note_text")
        buffer = fld.get_buffer()
        startIter = buffer.get_start_iter()
        endIter = buffer.get_end_iter()
        string = string +'\0' + buffer.get_text( startIter, endIter, True ) + '\0'
        self.records[self.curRecord][1] = self.EncodeRecord( string, self.pwDigest )

    def ClearButtons( self ) :
        updateBtn = self.wTree.get_widget("main_update_btn")
        updateBtn.set_sensitive( False )
        cancelBtn = self.wTree.get_widget("main_cancel_btn")
        cancelBtn.set_sensitive( False )

    def ActivateButtons( self ) :
        updateBtn = self.wTree.get_widget("main_update_btn")
        updateBtn.set_sensitive( True )
        cancelBtn = self.wTree.get_widget("main_cancel_btn")
        cancelBtn.set_sensitive( True )

    def SetFields( self, name, account, password, note ) :
        fld = self.wTree.get_widget("name_text")
        fld.set_text( name )
        fld = self.wTree.get_widget("account_text")
        fld.set_text( account )
        fld = self.wTree.get_widget("password_text")
        fld.set_text( password )
        fld = self.wTree.get_widget("note_text")
        buf = fld.get_buffer()
        buf.delete( buf.get_start_iter(), buf.get_end_iter() )
        iter = buf.get_start_iter()
        buf.insert( iter, note )
        fld.set_buffer( buf )

    def ClearFields( self ) :
        self.SetFields( '', '', '', '' )

    def OnRowNew( self, one ) :
        #print "New : ", one 

        if not self.CheckPassword() :
            return
            
        self.CheckCurrentRow( True )

        sel = self.nameView.get_selection()

        #print "Sel : ", sel

        selected = sel.get_selected()
        liststore = selected[0]

        iter = liststore.insert( 0 )
        self.records.insert( 0, ['','',''] )
        
        self.curRecord = 0

        path = liststore.get_path( iter ) 
        #print "path : ", path

        self.nameView.set_cursor_on_cell( path )

        cur = self.nameView.get_cursor()
        self.curRecord = cur[0][0]
        #print "cur : ", cur

        self.ClearButtons()
        self.ClearFields()
        self.recordChanged = 1

    def OnRowDelete( self, one ) :
        if not self.CheckPassword() :
            return
            
        if self.curRecord == None :
            return

        #curRecord = cursor[0][0]
        #print "Cur : ", self.curRecord

        sel = self.nameView.get_selection()

        #print "Sel : ", sel

        selected = sel.get_selected()
        liststore = selected[0]
        iter = selected[1]

        #print "Selelected : ", selected

        if self.recordChanged == 1 :
            self.recordChanged = 0
            self.ClearButtons()

        if self.curRecord != None :
            delete = self.show_warning_dialog( "Delete Row" )
            if delete == gtk.RESPONSE_OK :
                self.hasChanged = True
                del self.records[self.curRecord]
                liststore.remove( iter )
                self.ClearFields()
                self.recordChanged = 0

        #for record in self.records :
        #    print "Name : " , record[0]

        #cur = self.nameView.get_cursor()
        #self.curRecord = cur[0][0]

    def OnMainUpdateBtn( self, event ) :
        self.ClearButtons()

        if self.recordChanged == 1 :
            self.recordChanged = 0
            if self.curRecord != None :
                self.UpdateRecordFromRow()
                self.SetListItem( self.curRecord, self.records[self.curRecord][0] ) 

    def SetListItem( self, row, text ) :
        iter = self.nameList.get_iter( (row,) ) 
        self.nameList.set( iter, 0, text )

    def OnRowActivated( self, tree, event ) :
        #print "tree :", tree
        #print "event :", event

        if self.recordChanged == 1 :
            self.recordChanged = 0
            self.ClearButtons()
            if self.curRecord != None :
                save = self.show_warning_dialog( "Apply Changes ?" )
                if save == gtk.RESPONSE_OK :
                    self.UpdateRecordFromRow()
                    self.SetListItem( self.curRecord, records[curRecord][0] ) 

        cur = self.nameView.get_cursor()
        self.curRecord = cur[0][0]
        iter = self.nameList.get_iter( cur[0] ) 

        #print "CurRecord : ", self.curRecord

        if not self.CheckPassword() :
            return
            
        record = self.DecodeRecord( self.records[self.curRecord][1], 
                                    self.pwDigest )
        #print "record : ", record
        self.SetFields( self.records[self.curRecord][0], record[0],
                            record[1], record[2] )
        self.recordChanged = 0
        self.ClearButtons()
        
    def OnSelectRow( self, editing, user_param ) :
#        sel = self.nameView.get_selected()
#        sel = self.nameView.get_selection()
#        print "selected : ", sel
        cur = self.nameView.get_cursor()
        print "cursor : ", cur

    def AddListColumn(self, title, columnId, sort):
        """This function adds a column to the list view.
        First it create the gtk.TreeViewColumn and then set
        some needed properties"""

        column = gtk.TreeViewColumn(title, gtk.CellRendererText()
                                    , text=columnId)
        column.set_resizable(True)
        if sort :
            column.set_sort_order( gtk.SORT_DESCENDING )
            column.set_sort_column_id(columnId)
        else :
            column.set_clickable( False )

        self.nameView.append_column(column)

    def RecordString(self, record ) :
        i = 0
        while record[i] != '\0' :
            i = i + 1
        return record[:i]

    def readAlpha(self, f,n):
        retVal = f.read(n)
        return retVal

    def readShort(self,f):
        """Read unsigned 2 byte value from a file f."""
        (retVal,) = struct.unpack("H", f.read(2))
        return socket.ntohs( retVal )

    def readLong(self,f):
        """Read unsigned 4 byte value from a file f."""
        (retVal,) = struct.unpack("I", f.read(4))
        return socket.ntohl( retVal )

    def ReadKeyringHeader(self, f ) :
        f.seek(0)
        name = self.readAlpha(f,32)
        #print "Name : ", name
        dbAttr = self.readShort(f)
        version = self.readShort(f)
        #print "version : ", version
        if version != 4 :
            show_error_dialog("This hasn't been tested with pdb versions other than 4" )
        crDate = self.readLong(f)
        modDate = self.readLong(f)
        backupDate = self.readLong(f)
        modNumber = self.readLong(f)
        #print "Modification Number : ", modNumber
        self.appInfoID = self.readLong(f)
        #print "appinfo Number : ", self.appInfoID
        sortInfoID = self.readLong(f)
        dbType = self.readLong(f)
        creatID = self.readLong(f)
        idSeed = self.readLong(f)
        nextRecordListID = self.readLong(f)
        #print "Modification Number : ", nextRecordListID
        self.numRecords = self.readShort(f)
        self.recDirOffset = f.tell()
        #print "Records : ", self.numRecords
        #print "Dir offset : ", self.recDirOffset
        
        f.seek( self.recDirOffset )
        offset = self.readLong(f)
        f.seek( offset )
        self.salt = f.read( self.saltSize )
        # not valid for version 4 ?
        #self.iter = self.readShort(f)
        #self.cypher = self.readShort(f)
        self.hash = f.read( self.hashSize )
        
        #print "salt : ", str( self.salt ).encode( "hex" )
        #print "hash : ", str( self.hash ).encode( "hex" )
        
    def CheckPalmKeyringPasswd( self, password, updateDigest = True ) :
        m = hashlib.md5()
        m.update( self.salt )
        m.update( password )
        # pad with zeros
        m.update( ''.rjust( 64 - len( self.salt ) - len( password ), '\0' ) )
            
        digest = m.digest()

        #print "salt   : ", self.salt.encode( "hex" )
        #print "digest : ", digest.encode( "hex" )
        #print "hash   : ", self.hash.encode( "hex" )

        for i in range( 1 , len( self.hash )):
            if digest[i] != self.hash[i] :
                print "Wrong password"
                return False

        if updateDigest  :
            m = hashlib.md5( password )
            self.pwDigest = m.digest()

        password.zfill( len( password ))
        
        return True

    def ReadPalmKeyringDB(self, f ) :
        self.pwDigest = None
        
        self.ReadKeyringHeader( f )
        
        self.records = []
        # this skips the last record
        # this should not skip the first record 
        # it should just skip the hidden record
        for i in range ( 1, self.numRecords-1 ) :
            f.seek( i*8 + self.recDirOffset )
            recordOffset = self.readLong(f)
            recordAttr = f.read(1)
            recordID = f.read(3)
            #if not ( recordAttr[0] & 0x08 ):
            nextOffset = self.readLong(f)
            recordLen = nextOffset - recordOffset;
            self.records.append( self.ReadRecord( f, recordOffset, recordLen ))
            
        f.seek(( self.numRecords-1 )*8  + self.recDirOffset )
        recordOffset = self.readLong(f)
        recordAttr = f.read(1)

        #if not ( recordAttr[0] & 0x08 ):
        f.seek( 0, 2 )
        fileEnd = f.tell()
        recordLen = fileEnd - recordOffset
        self.records.append( self.ReadRecord( f, recordOffset, recordLen ))

        self.NameViewSort( self.nameList )
            
        for i in range ( 0, self.numRecords-1 ) :
            self.nameList.append( [self.records[i][0], ''] )

    def ReadRecord( self, f, offset, length ):
        f.seek( offset )
        record = f.read( length )
        recordName = self.RecordString( record )
        recordData = record[ len( recordName ) + 1:length]
        return [ recordName, recordData ] 

    def DecodeRecord( self, enc, digest ):
    
        k = pyDes.triple_des( digest, pyDes.ECB )
        mod = len( enc ) % 8

        if mod != 0 :
            enc = enc + ''.rjust( 8 - mod, '\0' )

        dec = k.decrypt( enc )
        
        #print "dec  : ", str( dec ).encode( "hex" )
        #print "Data : ", dec
        
        ret = []
        ret.append( self.RecordString( dec ))
        dec = dec[ len( ret[0] ) + 1: len(dec) ]
        ret.append( self.RecordString( dec ))
        dec = dec[ len( ret[1] ) + 1: len(dec) ]
        ret.append( self.RecordString( dec ))
        #print "ret : ", ret

        return ret

    def EncodeRecord( self, dec, digest ):
    
        k = pyDes.triple_des( digest, pyDes.ECB )
        mod = len( dec ) % 8

        # should add random data instead
        if mod != 0 :
            dec = dec + ''.rjust( 8 - mod, '\0' )

        enc = k.encrypt( dec )
        
        return enc

    def show_error_dialog(self, error_string):
	"""This Function is used to show an error dialog when an error occurs.
	error_string - The error string that will be displayed on the dialog.
	"""
	error_dlg = gtk.MessageDialog(type=gtk.MESSAGE_ERROR
				, message_format=error_string
				, buttons=gtk.BUTTONS_OK)
	error_dlg.run()
	error_dlg.destroy()

    def show_warning_dialog(self, msg_string):
	"""This Function is used to show an error dialog when an error occurs.
	msg_string - The error string that will be displayed on the dialog.
	"""
	msg_dlg = gtk.MessageDialog(type=gtk.MESSAGE_WARNING
				, message_format=msg_string
				, buttons=gtk.BUTTONS_OK_CANCEL )
	ret = msg_dlg.run()
	msg_dlg.destroy()
        return ret

start = Main()
gtk.main()
